import os
import time
from pathlib import Path
from typing import Iterator, List, Optional, TextIO

import click
from single_source import get_version

from autopr import config, database, github, repo, workdir
from autopr.util import CliException, error, is_debug, set_debug

__version__ = get_version(
    "auto-pr",
    Path(__file__).parent.parent,
)

DEFAULT_PUSH_DELAY = 30.0
WORKDIR: workdir.WorkDir


def main():
    try:
        cli()
    except CliException as e:
        error(f"Error: {e}")
        if is_debug():
            raise e
        else:
            exit(1)
    except KeyboardInterrupt:
        error("Aborted.")


def _ensure_set_up(cfg: config.Config, db: database.Database):
    if db.needs_pulling():
        raise CliException("No data found. Please run 'pull' first.")

    if len(cfg.update_command) == 0:
        raise CliException(
            "No update command found. Please set an update command in the config."
        )


@click.group("auto-pr")
@click.option(
    "-w",
    "--workdir",
    "wd_path",
    envvar="APR_WORKDIR",
    type=click.Path(
        exists=False, file_okay=False, dir_okay=True, writable=True, readable=True
    ),
    help="Working directory to store configuration and repositories",
)
@click.option(
    "--debug/--no-debug",
    envvar="APR_DEBUG",
    default=False,
    is_flag=True,
    help="Whether to enable debug mode or not",
)
@click.version_option(__version__, message="%(prog)s: %(version)s")
def cli(wd_path: str, debug: bool):
    global WORKDIR
    WORKDIR = workdir.get(wd_path)
    set_debug(debug)


@cli.command()
@click.option(
    "--api-key",
    envvar="APR_API_KEY",
    required=True,
    prompt="GitHub API key",
    hide_input=True,
    help="The GitHub API key to use, needs `repo` and `user->user:email` scope",
)
@click.option(
    "--ssh-key-file",
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, writable=False, readable=True
    ),
    required=True,
    prompt="Private SSH keyfile",
    help="Path to the SSH key to use when pushing to GitHub",
)
def init(api_key: str, ssh_key_file: str):
    """Initialise configuration and database"""
    credentials = config.Credentials(api_key=api_key, ssh_key_file=ssh_key_file)
    workdir.init(WORKDIR, credentials)


@cli.command()
@click.option(
    "--fetch-repo-list/--no-fetch-repo-list",
    default=True,
    is_flag=True,
    help="Whether to fetch the repo list from github or not",
)
@click.option(
    "--update-repos/--no-update-repos",
    default=False,
    is_flag=True,
    help="Whether to update the locally cloned repositories to latest changes",
)
@click.option(
    "--process-count",
    "-j",
    default=os.cpu_count() or 1,
    type=int,
    help="How many repositories to pull in parallel",
)
@click.option(
    "--use-global-git-config/--use-primary-email-git-config",
    default=False,
    is_flag=True,
    help="Whether to use the already globally set git config or the primary email of the authenticated Github user. If you have already pulled the repos locally, you also need to pass --update-repos to update the git config in the repos.",
)
def pull(
    fetch_repo_list: bool,
    update_repos: bool,
    process_count: int,
    use_global_git_config: bool,
):
    """Pull down repositories based on configuration"""
    cfg = workdir.read_config(WORKDIR)
    gh = github.create_github_client(cfg.credentials.api_key)
    user = github.get_user(gh, use_global_git_config)

    click.secho(f"Running under user '{user.name}' with email '{user.email}'")

    # get repositories
    db_old = workdir.read_database(WORKDIR)

    if fetch_repo_list:
        click.secho("Gathering repository list...")
        repositories = github.gather_repository_list(gh, cfg.repositories)

        # merge existing database with new one
        click.secho("Updating database")
        db_new = database.Database(user=user, repositories=repositories)
        db_old.merge_into(db_new)
        workdir.write_database(WORKDIR, db_old)
    else:
        click.secho("Not gathering repository list")
        repositories = db_old.repositories_to_process()

    # pull all repositories
    click.secho("Pulling repositories...")
    repo.pull_repositories_parallel(
        user,
        Path(cfg.credentials.ssh_key_file),
        WORKDIR.repos_dir,
        repositories,
        update_repos,
        process_count,
    )


@cli.command()
@click.option(
    "--pull-repos/--no-pull-repos",
    default=False,
    is_flag=True,
    help="Whether to pull repositories before testing",
)
def test(pull_repos: bool):
    """Check what expected diff will be for command execution"""
    cfg = workdir.read_config(WORKDIR)
    db = workdir.read_database(WORKDIR)
    _ensure_set_up(cfg, db)

    for repository in db.repositories_to_process():
        try:
            repo.reset_and_run_script(repository, db, cfg, WORKDIR, pull_repos)
            diff = repo.get_diff(WORKDIR.repos_dir, repository)
            click.secho(f"Diff for repository '{repository.name}':\n{diff}")
        except CliException as e:
            error(f"Error: {e}")

        if not click.confirm("Continue?"):
            return

        click.secho("\n")


@cli.command()
@click.option(
    "--pull-repos/--no-pull-repos",
    default=True,
    is_flag=True,
    help="Whether to pull repositories before running",
)
@click.option(
    "--push-delay",
    type=click.FloatRange(min=0.0, max=None, clamp=True),
    default=DEFAULT_PUSH_DELAY,
    help="Delay in seconds between pushing changes to repositories",
)
@click.option(
    "--api-key",
    envvar="APR_API_KEY",
    required=False,
    hide_input=True,
    help="The GitHub API key to use if not statically configured",
)
def run(pull_repos: bool, push_delay: Optional[float], api_key: Optional[str]):
    """Run update logic and create pull requests if changes made"""
    cfg = workdir.read_config(WORKDIR)
    if api_key is not None:
        cfg.credentials.api_key = api_key
    db = workdir.read_database(WORKDIR)
    _ensure_set_up(cfg, db)
    gh = github.create_github_client(cfg.credentials.api_key)

    repositories = db.repositories_to_process()

    change_pushed = False
    for i, repository in enumerate(repositories, start=1):
        if change_pushed and push_delay is not None:
            click.secho(f"Sleeping for {push_delay} seconds...")
            time.sleep(push_delay)

        click.secho(
            f"[{i}/{len(repositories)}] Updating '{repository.name}'", bold=True
        )

        try:
            repo.reset_and_run_script(repository, db, cfg, WORKDIR, pull_repos)
            change_pushed = repo.push_changes(repository, db, cfg, gh, WORKDIR)
        except CliException as e:
            error(f"Error: {e}")

    click.secho(f"Done!", bold=True)


@cli.group()
def reset():
    """Commands for resetting repos to allow for reruns"""
    pass


@reset.command(name="all")
def reset_all():
    """Mark all mapped repositories as not done"""
    db = workdir.read_database(WORKDIR)
    db.reset_all()
    workdir.write_database(WORKDIR, db)
    click.secho("Repositories marked as not done")


@reset.command(
    name="from", help="reset using a file listing of repos written as <owner>/<name>"
)
@click.argument("file", type=click.File("r"))
def reset_from(file: TextIO):
    repos: Iterator[str] = map(lambda l: l.strip(), file.readlines())
    db = workdir.read_database(WORKDIR)
    db.reset_from(repos)
    workdir.write_database(WORKDIR, db)


def _print_repository_list(
    title: str, repositories: List[database.Repository], total: int
):
    click.secho(f"{title} [{len(repositories)}/{total}]:", bold=True)
    for repository in repositories:
        link_str = ""
        if repository.existing_pr is not None:
            link_str = (
                f": https://github.com/{repository.full_name}"
                f"/pull/{repository.existing_pr}"
            )
        click.secho(f"-   {repository.name}{link_str}")


@cli.command()
@click.option(
    "--exclude-missing/--include-missing",
    default=False,
    is_flag=True,
    help="Whether the `Missing PRs` section should be excluded from status.",
)
def status(exclude_missing: bool):
    cfg = workdir.read_config(WORKDIR)
    db = workdir.read_database(WORKDIR)
    gh = github.create_github_client(cfg.credentials.api_key)

    click.secho("Collecting data...")

    if len(db.repositories) == 0:
        error("No repositories in database.")
        return

    pr_missing = []
    pr_merged = []
    pr_open = []
    pr_closed = []

    # group repositories by PR state
    for repository in db.repositories:
        if repository.existing_pr is None:
            pr_missing.append(repository)
            continue

        pull_request = github.get_pull_request(gh, repository)
        if pull_request.merged:
            pr_merged.append(repository)
        elif pull_request.state == github.PullRequestState.OPEN.value:
            pr_open.append(repository)
        elif pull_request.state == github.PullRequestState.CLOSED.value:
            pr_closed.append(repository)

    total = len(db.repositories)
    _print_repository_list("Merged PRs", pr_merged, total)
    _print_repository_list("Closed PRs", pr_closed, total)
    _print_repository_list("Open PRs", pr_open, total)
    if not exclude_missing:
        _print_repository_list("Missing PRs", pr_missing, total)


def _set_all_pull_requests_state(state: github.PullRequestState):
    cfg = workdir.read_config(WORKDIR)
    db = workdir.read_database(WORKDIR)
    gh = github.create_github_client(cfg.credentials.api_key)

    for repository in db.repositories:
        if repository.existing_pr is not None:
            try:
                github.set_pull_request_state(gh, repository, state)
                click.secho(
                    f"Updated {repository.name} pull request state to {state.value}"
                )
            except ValueError as e:
                click.secho(f"{e}")


@cli.command()
def close():
    """Close all open pull requests"""
    _set_all_pull_requests_state(github.PullRequestState.CLOSED)
    click.secho("Finished closing all open pull requests")


@cli.command()
def reopen():
    """Reopen all un-merged pull requests"""
    _set_all_pull_requests_state(github.PullRequestState.OPEN)
    click.secho("Finished reopening all closed unmerged pull requests")


if __name__ == "__main__":
    main()



import io
import os
import shutil
import subprocess
import sys
from multiprocessing import Pool
from pathlib import Path
from typing import IO, Dict, List, Optional

import click
from github import Github

from autopr import config, database, github, util
from autopr.database import Repository
from autopr.util import CliException, error
from autopr.workdir import WorkDir, write_database


def _pull_repository_task(
    user: database.GitUser,
    ssh_key_file: Path,
    repos_dir: Path,
    repository: database.Repository,
    update_repo_if_exists: bool,
) -> None:
    try:
        output_buffer = io.StringIO()

        try:
            pull_repository(
                user,
                ssh_key_file,
                repos_dir,
                repository,
                update_repo_if_exists,
                out=output_buffer,
            )
        except CliException as e:
            error(f"Error: {e}", file=output_buffer)

        click.echo(output_buffer.getvalue(), nl=False)
        output_buffer.close()
    except KeyboardInterrupt:
        pass  # will be handled on main thread


def pull_repositories_parallel(
    user: database.GitUser,
    ssh_key_file: Path,
    repos_dir: Path,
    repositories: List[database.Repository],
    update_repos: bool,
    process_count: int,
):
    parameters = []
    for repository in repositories:
        parameters.append(
            (
                user,
                ssh_key_file,
                repos_dir,
                repository,
                update_repos,
            )
        )

    with Pool(processes=process_count) as pool:
        pool.starmap(_pull_repository_task, parameters)


def pull_repository(
    user: database.GitUser,
    ssh_key_file: Path,
    repos_dir: Path,
    repository: database.Repository,
    update_repo_if_exists: bool,
    out: IO[str] = sys.stdout,
) -> None:
    repo_dir = repos_dir / repository.name
    repo_exists = repo_dir.exists()

    if not update_repo_if_exists and repo_exists:
        click.echo(f"Repository '{repository.name}' already exists", file=out)
        return

    click.echo(f"Pulling repository '{repository.name}':", file=out)

    pull_failed = False
    if repo_exists:
        click.echo(f"  - Checking out branch '{repository.default_branch}'", file=out)
        _git_checkout(repo_dir, repository.default_branch)

        click.echo("  - Pulling latest changes", file=out)
        try:
            _git_pull(repo_dir)
        except CliException as e:
            util.debug(f"Failed to pull: {e}")
            pull_failed = True

    if pull_failed:
        click.echo("  - Pull failed; deleting repo", file=out)
        shutil.rmtree(repo_dir)

    if not repo_exists or pull_failed:
        click.echo(f"  - Cloning branch '{repository.default_branch}'", file=out)
        _git_shallow_clone(
            ssh_key_file, repo_dir, repository.ssh_url, repository.default_branch
        )

        click.echo("  - Setting user and email", file=out)
    _git_config(repo_dir, "user.name", user.name)
    _git_config(repo_dir, "user.email", user.email)


def prepare_repository(repos_dir: Path, repository: database.Repository, branch: str):
    repo_dir = repos_dir / repository.name

    click.echo(f"Resetting repository '{repository.name}':")

    click.echo("  - Resetting changes")
    _git_reset_hard(repo_dir)

    click.echo(f"  - Checking out default branch '{repository.default_branch}'")
    _git_checkout(repo_dir, repository.default_branch)

    click.echo(f"  - Creating or resetting branch '{branch}'")
    _git_branch_checkout_reset(repo_dir, branch)


def run_update_command(
    repos_dir: Path, repository: database.Repository, command: List[str]
):
    repo_dir = repos_dir / repository.name

    click.echo(f"Running update command for repository '{repository.name}':")

    cwd = os.getcwd()
    os.chdir(repo_dir)
    run_cmd(command)
    os.chdir(cwd)

    _git_add_all(repo_dir)


def get_diff(repos_dir: Path, repository: database.Repository):
    repo_dir = repos_dir / repository.name
    return _git_staged_diff(repo_dir)


def commit_and_push_changes(
    ssh_key_file: Path,
    repos_dir: Path,
    repository: database.Repository,
    branch: str,
    message: str,
) -> bool:
    repo_dir = repos_dir / repository.name
    if _git_staged_diff(repo_dir) == "":
        click.echo("  - No changes")
        return False

    click.echo("  - Committing and pushing changes")
    _git_commit(repo_dir, message)

    force_push = repository.existing_pr is not None
    _git_push(ssh_key_file, repo_dir, branch, force_push)

    return True


def run_cmd(cmd: List[str], additional_env: Optional[Dict[str, str]] = None) -> str:
    env = None
    if additional_env:
        env = os.environ.copy()
        env.update(additional_env)

    try:
        util.debug(f"Running: {' '.join(cmd)}")
        return subprocess.check_output(
            cmd,
            stderr=subprocess.STDOUT,
            env=env,
        ).decode()
    except subprocess.CalledProcessError as exc:
        raise CliException(
            f"Command {' '.join(cmd)} failed (code: {exc.returncode}):\n{exc.output.decode()}"
        )


def _get_git_ssh_command(ssh_key_file: Path) -> str:
    return f"ssh -i {ssh_key_file} -o IdentitiesOnly=yes"


def _git_shallow_clone(
    ssh_key_file: Path, repo_dir: Path, ssh_url: str, branch: str
) -> None:
    git_ssh_command = _get_git_ssh_command(ssh_key_file)
    command = [
        "git",
        "clone",
        "--depth",
        "1",
        ssh_url,
        f"{repo_dir}",
        "--branch",
        branch,
    ]

    run_cmd(command, additional_env={"GIT_SSH_COMMAND": git_ssh_command})


def _git_checkout(repo_dir: Path, branch: str) -> None:
    run_cmd(["git", "-C", f"{repo_dir}", "checkout", branch])


def _git_branch_checkout_reset(repo_dir: Path, branch: str) -> None:
    run_cmd(["git", "-C", f"{repo_dir}", "checkout", "-B", branch])


def _git_reset_hard(repo_dir: Path) -> None:
    run_cmd(["git", "-C", f"{repo_dir}", "reset", "--hard"])


def _git_pull(repo_dir: Path) -> None:
    run_cmd(["git", "-C", f"{repo_dir}", "pull", "--depth", "1"])


def _git_config(repo_dir: Path, key: str, value: str) -> None:
    run_cmd(["git", "-C", f"{repo_dir}", "config", key, value])


def _git_get_global_config(key: str) -> str:
    return run_cmd(["git", "config", "--global", key])


def _git_staged_diff(repo_dir: Path) -> str:
    return run_cmd(
        ["git", "-c", "color.ui=always", "-C", f"{repo_dir}", "diff", "--staged"]
    )


def _git_add_all(repo_dir: Path) -> None:
    run_cmd(["git", "-C", f"{repo_dir}", "add", "--all"])


def _git_commit(repo_dir: Path, message) -> None:
    run_cmd(["git", "-C", f"{repo_dir}", "commit", "-m", message])


def _git_push(ssh_key_file: Path, repo_dir: Path, branch: str, force: bool) -> None:
    git_ssh_command = _get_git_ssh_command(ssh_key_file)
    cmd = ["git", "-C", f"{repo_dir}", "push", "-u", "origin", branch]
    if force:
        cmd.append("--force")

    run_cmd(cmd, additional_env={"GIT_SSH_COMMAND": git_ssh_command})


def reset_and_run_script(
    repository: Repository,
    db: database.Database,
    cfg: config.Config,
    workdir: WorkDir,
    pull_repo: bool,
):
    if db.user is None:
        raise Exception(
            "db.user is None - please report at github.com/getyourguide/auto-pr"
        )

    if pull_repo:
        pull_repository(
            db.user,
            Path(cfg.credentials.ssh_key_file),
            workdir.repos_dir,
            repository,
            True,
        )

    # reset repo and check out branch
    prepare_repository(workdir.repos_dir, repository, cfg.pr.branch)
    run_update_command(workdir.repos_dir, repository, cfg.update_command)


def push_changes(
    repository: Repository,
    db: database.Database,
    cfg: config.Config,
    gh: Github,
    workdir: WorkDir,
) -> bool:
    updated = commit_and_push_changes(
        Path(cfg.credentials.ssh_key_file),
        workdir.repos_dir,
        repository,
        cfg.pr.branch,
        cfg.pr.message,
    )

    if not updated:
        click.secho("  - Nothing updated")
        _mark_repository_as_done(repository, db, workdir)
        return False

    if repository.existing_pr:
        pull_request = github.get_pull_request(gh, repository)
        if (
            not pull_request.merged
            and pull_request.state != github.PullRequestState.CLOSED.value
        ):
            click.secho(f"  - Pull request: {pull_request.html_url}")
            _mark_repository_as_done(repository, db, workdir)
            return False

    pull_request = github.create_pr(gh, repository, cfg.pr)
    repository.existing_pr = pull_request.number

    click.secho(f"  - Pull request: {pull_request.html_url}")

    # persist database to be able to continue from there
    _mark_repository_as_done(repository, db, workdir)
    click.secho(f"Done updating repository '{repository.name}'")

    return True


def _mark_repository_as_done(
    repository: Repository, db: database.Database, workdir: WorkDir
):
    repository.done = True
    write_database(workdir, db)



import re
from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Tuple

from github import Github
from github.PullRequest import PullRequest

from autopr import config, database
from autopr.repo import _git_get_global_config
from autopr.util import CliException


@dataclass
class FilterInfo:
    owner: str
    name: str
    public: bool
    archived: bool


class PullRequestState(Enum):
    OPEN = "open"
    CLOSED = "closed"


def create_github_client(api_key: str) -> Github:
    gh = Github(api_key, per_page=150)
    return gh


def get_user(gh: Github, use_global_git_config: bool = False) -> database.GitUser:
    gh_user = gh.get_user()

    login = gh_user.login  # need to do this first to trigger lazy loading
    name = gh_user.name

    if use_global_git_config:
        primary_email = _git_get_global_config("user.email").strip()
        name = _git_get_global_config("user.name").strip()
        if not name or not primary_email:
            raise CliException(
                "You don't have a globally set Git config. "
                "Please set your config with `git config --global user.email <EMAIL>` or `git config --global user.name <NAME>`"
            )
    else:
        emails = gh_user.get_emails()
        primary_email = None
        for email in emails:
            if email.primary:
                primary_email = email.email
                break

        user_name = name or login
        if user_name is None or primary_email is None:
            raise CliException(
                "Please provide an API key with access to the user name and email address."
            )

    user = database.GitUser(name=name or login, email=primary_email)
    return user


def create_pr(
    gh: Github, repository: database.Repository, pr_template: config.PrTemplate
) -> PullRequest:
    gh_repo = gh.get_repo(repository.full_name)
    pull_request = gh_repo.create_pull(
        base=repository.default_branch,
        head=pr_template.branch,
        title=pr_template.title,
        body=pr_template.body,
        maintainer_can_modify=True,
        draft=pr_template.draft,
    )
    return pull_request


def get_pull_request(gh: Github, repository: database.Repository) -> PullRequest:
    gh_repo = gh.get_repo(repository.full_name)
    return gh_repo.get_pull(repository.existing_pr)


def set_pull_request_state(
    gh: Github, repository: database.Repository, state: PullRequestState
):
    if repository.existing_pr is None:
        raise ValueError(f"No existing pull request for {repository.name}")

    gh_repo = gh.get_repo(repository.full_name)
    pull_request = gh_repo.get_pull(repository.existing_pr)

    if pull_request.merged:
        raise ValueError(
            f"Pull request already merged for {repository.name} ({pull_request.html_url})"
        )

    pull_request.edit(state=state.value)


def gather_repository_list(
    gh: Github, filters: List[config.Filter]
) -> List[database.Repository]:
    all_repositories = _list_all_repositories(gh)
    filtered_repositories = _apply_filters(all_repositories, filters)
    return filtered_repositories


def _list_all_repositories(
    gh: Github,
) -> List[Tuple[FilterInfo, database.Repository]]:
    repository_list = []
    gh_repo_list = gh.get_user().get_repos()
    for gh_repo in gh_repo_list:
        repository = database.Repository(
            owner=gh_repo.owner.login,
            name=gh_repo.name,
            ssh_url=gh_repo.ssh_url,
            default_branch=gh_repo.default_branch,
        )
        filter_info = FilterInfo(
            owner=gh_repo.owner.login,
            name=gh_repo.name,
            public=not gh_repo.private,
            archived=gh_repo.archived,
        )
        repository_list.append((filter_info, repository))

    return repository_list


def _apply_filters(
    all_repositories: List[Tuple[FilterInfo, database.Repository]],
    filters: List[config.Filter],
) -> List[database.Repository]:
    # need to create a dict with (owner, name) as the key, as the type is not hashable
    selected_repositories: Dict[
        Tuple[str, str], Tuple[FilterInfo, database.Repository]
    ] = {}

    if len(filters) == 0:
        return [repository for _filter_info, repository in all_repositories]

    for filter in filters:
        if filter.mode == config.FILTER_MODE_ADD:
            added_repositories = {
                (repository.owner, repository.name): (filter_info, repository)
                for filter_info, repository in all_repositories
                if _filter_matches(filter, filter_info)
            }
            selected_repositories.update(added_repositories)
        elif filter.mode == config.FILTER_MODE_REMOVE:
            selected_repositories = {
                (repository.owner, repository.name): (filter_info, repository)
                for filter_info, repository in selected_repositories.values()
                if not _filter_matches(filter, filter_info)
            }
        else:
            raise CliException(f"Unsupported filter.mode passed: {filter.mode}")

    return [repository for _filter_info, repository in selected_repositories.values()]


def _filter_matches(filter: config.Filter, filter_info: FilterInfo):
    if filter.public is not None:
        if filter_info.public != filter.public:
            return False

    if filter.archived is not None:
        if filter_info.archived != filter.archived:
            return False

    if filter.match_name is not None:
        if all(
            not re.fullmatch(pattern, filter_info.name) for pattern in filter.match_name
        ):
            return False

    if filter.match_owner is not None:
        if not re.fullmatch(filter.match_owner, filter_info.owner):
            return False

    return True



from dataclasses import dataclass, field
from typing import Dict, Iterator, List, Optional

import marshmallow_dataclass


@dataclass
class Repository:
    owner: str
    name: str
    ssh_url: str
    default_branch: str
    existing_pr: Optional[int] = None
    removed: bool = False  # true if the repo is not in the config's filters anymore, but still has a PR open
    done: bool = False  # true if a PR has been opened and 'reset' was not called

    @property
    def full_name(self) -> str:
        return f"{self.owner}/{self.name}"


repository_schema = marshmallow_dataclass.class_schema(Repository)()


@dataclass
class GitUser:
    name: str
    email: str


GIT_USER_SCHEMA = marshmallow_dataclass.class_schema(GitUser)()


@dataclass
class Database:
    user: Optional[GitUser] = None
    repositories: List[Repository] = field(
        default_factory=list
    )  # is equal to assigning []

    def repositories_to_process(self) -> List[Repository]:
        """Get all repositories filtering out done and removed"""
        return [
            repo for repo in self.repositories if not repo.removed and not repo.done
        ]

    def needs_pulling(self) -> bool:
        return self.user is None

    def merge_into(self, from_db: "Database") -> None:
        self.user = from_db.user

        # add repositories that are new
        existing_repos = set(
            (repository.owner, repository.name) for repository in self.repositories
        )
        for repository in from_db.repositories:
            if (repository.owner, repository.name) not in existing_repos:
                self.repositories.append(repository)

        # mark repositories that are gone as removed
        new_repos = set(
            (repository.owner, repository.name) for repository in from_db.repositories
        )
        for repository in self.repositories:
            if (repository.owner, repository.name) not in new_repos:
                repository.removed = True

    def reset_from(self, selected_repos: Iterator[str]):
        resets: Dict[str, bool] = {name: False for name in selected_repos}
        for repository in self.repositories:
            repo_id = f"{repository.owner}/{repository.name}"
            if repo_id in resets:
                print(f"{repo_id} was reset")
                resets[repo_id] = True
                repository.done = False

        for name, done in resets.items():
            if not done:
                print(f"{name} was not in the database")

    def reset_all(self) -> None:
        for repository in self.repositories:
            repository.done = False


DATABASE_SCHEMA = marshmallow_dataclass.class_schema(Database)()



import json
from pathlib import Path

import yaml
from marshmallow import ValidationError

from autopr import config, database
from autopr.util import CliException, warning

CONFIG_FILE_NAME = "config.yaml"
DB_FILE_NAME = "db.json"
REPOS_DIR_NAME = "repos"


class WorkDir:
    location: Path

    def __init__(self, location: Path):
        self.location = location

    @property
    def config_file(self) -> Path:
        return self.location / CONFIG_FILE_NAME

    @property
    def database_file(self) -> Path:
        return self.location / DB_FILE_NAME

    @property
    def repos_dir(self) -> Path:
        return self.location / REPOS_DIR_NAME


def init(wd: WorkDir, credentials: config.Credentials):
    # create work dir and repos dir
    wd.repos_dir.mkdir(parents=True, exist_ok=True)

    # create default config
    if not wd.config_file.exists():
        pr = config.PrTemplate()
        cfg = config.Config(credentials=credentials, pr=pr)
        write_config(wd, cfg)
    else:
        warning("config file exists - not overriding")

    # create empty database
    if not wd.database_file.exists():
        db = database.Database()
        write_database(wd, db)
    else:
        warning("database file exists - not overriding")


def write_config(wd: WorkDir, cfg: config.Config):
    # load config file
    try:
        data = config.CONFIG_SCHEMA.dump(cfg)
        with open(wd.config_file, "w") as config_file:
            yaml.dump(data, config_file, default_flow_style=False)
    except IOError as e:
        raise CliException(f"Failed to write config file: {e}")


def read_config(wd: WorkDir) -> config.Config:
    # load config file
    try:
        with open(wd.config_file) as config_file:
            config_dict = yaml.safe_load(config_file)
    except IOError as e:
        raise CliException(f"Failed to read config file: {e}")
    except yaml.YAMLError as e:
        raise CliException(f"Failed to parse config: {e}")

    # parse config data
    try:
        return config.CONFIG_SCHEMA.load(config_dict)
    except ValidationError as err:
        raise CliException(f"Failed to deserialize config: {err.messages}")


def write_database(wd: WorkDir, db: database.Database):
    # load database file
    try:
        data = database.DATABASE_SCHEMA.dump(db)
        with open(wd.database_file, "w") as database_file:
            json.dump(data, database_file, indent=4, sort_keys=True)
    except IOError as e:
        raise CliException(f"Failed to write database file: {e}")


def read_database(wd: WorkDir) -> database.Database:
    if not wd.database_file.exists():
        db = database.Database()
        return db

    # load database file
    try:
        with open(wd.database_file) as database_file:
            database_dict = json.load(database_file)
    except IOError as e:
        raise CliException(f"Failed to read database file: {e}")
    except json.JSONDecodeError as e:
        raise CliException(f"Failed to parse database: {e}")

    # parse database data
    try:
        return database.DATABASE_SCHEMA.load(database_dict)
    except ValidationError as err:
        raise CliException(f"Failed to deserialize database: {err.messages}")


def get(wd_path: str) -> WorkDir:
    if wd_path:
        workdir_path = Path(wd_path)
    else:
        workdir_path = Path.cwd()

    return WorkDir(workdir_path)



from dataclasses import dataclass, field
from typing import List, Optional

import marshmallow_dataclass

DEFAULT_PR_TITLE = "Automatically generated PR"
DEFAULT_PR_MESSAGE = "Automatically generated commit"
DEFAULT_PR_BODY = "This is an automatically generated PR"
DEFAULT_PR_BRANCH = "autopr"

FILTER_MODE_ADD = "add"
FILTER_MODE_REMOVE = "remove"

FILTER_VISIBILITY_PUBLIC = "public"
FILTER_VISIBILITY_PRIVATE = "private"


@dataclass
class Credentials:
    api_key: str
    ssh_key_file: str


CREDENTIALS_SCHEMA = marshmallow_dataclass.class_schema(Credentials)()


@dataclass
class PrTemplate:
    title: str = DEFAULT_PR_TITLE
    message: str = DEFAULT_PR_MESSAGE
    branch: str = DEFAULT_PR_BRANCH
    body: str = DEFAULT_PR_BODY
    draft: bool = False


PR_TEMPLATE_SCHEMA = marshmallow_dataclass.class_schema(PrTemplate)()


@dataclass
class Filter:
    mode: str  # either 'add' or 'remove'
    public: Optional[bool] = None  # maps to the visibility of the repo
    archived: Optional[bool] = None
    match_name: Optional[
        List[str]
    ] = None  # a list of regex that are applied to the names, must match one
    match_owner: Optional[
        str
    ] = None  # a regex that is applied to the owners, must match one


FILTERS_SCHEMA = marshmallow_dataclass.class_schema(Filter)()


@dataclass
class Config:
    credentials: Credentials
    pr: PrTemplate
    repositories: List[Filter] = field(default_factory=list)  # is equal to assigning []
    update_command: List[str] = field(default_factory=list)


CONFIG_SCHEMA = marshmallow_dataclass.class_schema(Config)()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['autopr']

package_data = \
{'': ['*']}

install_requires = \
['PyGithub==2.6.1',
 'PyYAML==6.0.2',
 'click==8.1.7',
 'cryptography>=45.0.0',
 'marshmallow-dataclass==8.6.1',
 'marshmallow==3.21.3',
 'single-source==0.4.0']

entry_points = \
{'console_scripts': ['auto-pr = autopr:main']}

setup_kwargs = {
    'name': 'auto-pr',
    'version': '1.1.0',
    'description': 'Perform bulk updates across repositories',
    'long_description': '<img width="128" src="https://github.com/getyourguide/auto-pr/raw/master/img/logo.svg" alt="auto-pr logo" />\n\n![CI](https://github.com/getyourguide/auto-pr/workflows/CI/badge.svg)\n[![Publish](https://github.com/getyourguide/auto-pr/actions/workflows/publish.yml/badge.svg)](https://github.com/getyourguide/auto-pr/actions/workflows/publish.yml)\n[![PyPI version](https://badge.fury.io/py/auto-pr.svg)](https://badge.fury.io/py/auto-pr)\n![PyPI downloads](https://img.shields.io/pypi/dm/auto-pr)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n# auto-pr\n\nA command line tool to perform bulk updates across multiple GitHub repositories.\n\n## How to install\n\nWith [pipx](https://pipxproject.github.io/pipx/) (recommended):\n\n```bash\npipx install auto-pr\n```\nWith pip:\n\n```bash\npip install auto-pr\n```\n\n## Usage\n\n[![Usage](https://github.com/getyourguide/auto-pr/raw/master/img/workflow.svg)](https://github.com/getyourguide/auto-pr/raw/master/img/workflow.svg)\n\n### Init\n\nFirst initialise the project directory by running the `init` command within an empty directory.\n\n```bash\nauto-pr init --api-key=<github_token> --ssh-key-file=<path-to-ssh-key>\n```\n\nWhere `<github_token>` is a GitHub [personal access token](https://github.com/settings/tokens) which has `repo` and `user:user:email` scope.\n\nNext modify the generated `config.yaml` file with your desired configurations.\n\n```yaml\ncredentials:\n  api_key: <github_token>\n  ssh_key_file: /path/to/ssh/key/to/push/.ssh/id_rsa\npr:\n  body: >\n    Body of the PR that will be generated\n\n    Can be multi-line :)\n  branch: auto-pr # The branch name to use when making changes\n  message: Replace default pipelines with modules # Commit message\n  title: \'My awesome change\' # Title of the PR\n  draft: True # Whether to create the PR as a draft\nrepositories: # Rules that define what repos to update\n  - mode: add\n    match_owner: <org/user>\nupdate_command:\n  - touch\n  - my-file\n```\n\nIf you wish to keep your API Key outside of `config.yaml`, set the env var `APR_API_KEY` with your GitHub Token\n\n### Repositories\n\nYou can define the list of repositories to pull and build into the database to update using a list of rules.\n\n-   `mode` - either `add` or `remove` - used to either match or negate\n-   `public` (optional) - pull only public or private, leave out for both\n-   `archived` (optional) -  archived or non-archived, leave out for both\n-   `match_owner` (optional) - the owner or user to pull\n-   `match_name` (optional) - a list of regular expressions to match against to pull\n\nThe flags of the filter rules are optional not specifying will run the command on all repositories that the token has access too.\n\n### Update Command\n\nThis is the list containing the command to be executed along with the arguments passed to it. It will be executed from\nthe root of each repository that is processed.\n\nIf an error occurs during the execution it will be displayed in the output but will not halt the execution.\n\nSee [example commands](docs/examples.md#commands)\n\n### Pull\n\nAfter you have configured the project you can now pull the repositories down that match your rules.\n\n```bash\nauto-pr pull\n```\n\nThis will generate a `db.json` file within your workdir containing a list of mapped repositories and their state.\n\nThis command can be run multiple times, if there are new matching repositories found they will be merged into the existing database.\n\nIf you would like to use your globally set config, you can pass the option `--use-global-git-config` when pulling the repos. If you had already pulled the repos before this and you would like to change the config for those repos, you would also need to pass `--update-repos` alongside the global-git-config option when pulling.\n\n### Test\n\nOnce the `pull` command has finished setting up the work directory you can now run test to check what the changes that will be made by the script will yield.\n\n### Run\n\nWhen you\'re confident with the changes output from the `test` command you can finally execute `run`.\n\n```bash\nauto-pr run\n```\n\nThis will perform the changes to a branch on the locally cloned repository and push the branch upstream with the information you provided within `config.yaml`.\n\nBy default, the commits will be associated with your primary email and name, which were set on the repo level for those repos when you ran `auto-pr pull`. If you would like to use your global git config for the repos that you already pulled, you need to run pull again with:\n\n```\nauto-pr pull --update-repos --use-global-git-config\n```\n\nSee `--help` for more information about other commands and their  usage.\n\n### Reset\nYou can reset the list of repos in `db.json` using `auto-pr reset all`, or `auto-pr reset from FILE`\n\nWhen using `auto-pr reset from FILE`, the list of repos should be provided as a newline separated list of repos like `<owner>/<name>`, e.g:\n\n```text\ngetyourguide/test\ngetyourguide/auto-pr\n```\n\n## Security\n\nFor sensitive security matters please contact [security@getyourguide.com](mailto:security@getyourguide.com).\n\n## Legal\n\nCopyright 2021 GetYourGuide GmbH.\n\nauto-pr is licensed under the Apache License, Version 2.0. See [LICENSE](LICENSE) for the full text.\n',
    'author': 'GetYourGuide GmbH',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/getyourguide/auto-pr',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9.2,<4.0',
}


setup(**setup_kwargs)



import click

DEBUG: bool = True


class CliException(Exception):
    pass


def set_debug(debug: bool) -> None:
    global DEBUG
    DEBUG = debug


def is_debug() -> bool:
    return DEBUG


def debug(msg: str) -> None:
    if not is_debug():
        return

    click.secho(msg)


def warning(msg: str) -> None:
    click.secho(msg, fg="yellow", err=True)


def error(msg: str, **kwargs) -> None:
    click.secho(msg, fg="red", err=True, **kwargs)
